(*----------------------------------------------------------------------------*
 *  Direct3D FontDot3 rendering class                                         *
 *  demostrating how to create and render Dot product (Dot3) bumpmapped font. *
 *  Font normal map generated with D3DX8.1 function D3DXComputeNormalMap,     *
 *  height map is generated from font bitmap by averaging intensivity of 3x3  *
 *  square of pixels.                                                         *
 *                                                                            *
 *  Note: This class requires DirectX 8.1 compatible headers to compile !!!   *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 29-May-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit D3DFontDot3;

interface

{$I DirectX.inc}

uses
  Windows, D3DFont, {$I UseD3D8.inc}, D3DX8, DXUtil;

type
  TD3DFontDot3 = class(CD3DFont)
  private
  public
    fOnlyDot3: Boolean;
    fModulate2X: Boolean;
    fBlendToBlack: Boolean;
    fNormal: TD3DXVector3;
    // This function is not vistual so be carefull then calling it...
    function InitDeviceObjects(pd3dDevice: IDirect3DDevice8): HRESULT;
    function RestoreDeviceObjects: HRESULT;
    function DeleteDeviceObjects: HRESULT;
    function DrawText(x, y: Single; dwColor: DWORD;
      strText: PChar; dwFlags: DWORD = 0): HRESULT;
    function SetLightNormal(n: TD3DXVector3): HRESULT;
    destructor Destroy; override;
  end;

implementation

const
  MAX_NUM_VERTICES = 50*6;

type
  FONT2DVERTEX = packed record
    p: TD3DXVector4;
    color: DWORD;
    tu, tv: Single;
  end;

const
  D3DFVF_FONT2DVERTEX = (D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX1);

function InitFont2DVertex(const p: TD3DXVector4; color: TD3DColor;
  tu, tv: Single): FONT2DVERTEX;
var
  v: FONT2DVERTEX;
begin
  v.p := p;   v.color := color;   v.tu := tu;   v.tv := tv;
  Result:= v;
end;

const
  KERNEL_SIZE = 3;

//-----------------------------------------------------------------------------
// Name: VectortoRGBA()
// Desc: Turns a normalized vector into RGBA form. Used to encode vectors into
//       a height map.
//-----------------------------------------------------------------------------
function VectortoRGBA(const v: TD3DXVector3; fHeight: Single): DWord;
var
  r, g, b, a: DWord;
begin
  r := Round(127.0 * v.x + 128.0);
  g := Round(127.0 * v.y + 128.0);
  b := Round(127.0 * v.z + 128.0);
  a := Round(255.0 * fHeight);

  Result:= ((a shl 24) + (r shl 16) + (g shl 8) + (b shl 0));
end;

{ TD3DFontDot3 }

function TD3DFontDot3.InitDeviceObjects(
  pd3dDevice: IDirect3DDevice8): HRESULT;
var
  m_pNormTexture: IDirect3DTexture8;
  d3dlr, d3dlrHeight: TD3DLockedRect;
  pDstData, pDstMax, pDstRow: PByte;
  pDst16: PWord;
  x, y, i, j: Integer;
  HeightTexture: IDirect3DTexture8; // temporary texture for height data
  pSwzData: PDWORD; // original data but in swizzled format
  pIntData: PDWORD; // "intensivity" data (in swizzled format)
  Acc: Integer;
begin
  // Call "inherited" function
  inherited InitDeviceObjects(pd3dDevice);

  // Lock the surface and calculate Dot3 normal map from it
  m_pTexture.LockRect(0, d3dlr, nil, 0);
  pDstData:= d3dlr.pBits; // pDstData - is 16bit A4R4G4B4 surface
  pDstMax:= PByte(Integer(pDstData) + Integer(m_dwTexHeight)*d3dlr.Pitch);

  // Create and lock Height texture
  Result:= m_pd3dDevice.CreateTexture(m_dwTexWidth, m_dwTexHeight, 1,
                                      0, D3DFMT_A8R8G8B8,
                                      D3DPOOL_MANAGED, HeightTexture);
  if FAILED(Result) then Exit;
  HeightTexture.LockRect(0, d3dlrHeight, nil, 0);
  pSwzData:= d3dlrHeight.pBits; // pDstData - is 16bit A4R4G4B4 surface

  // Calculate per-pixel intensivity of some area
  for y:= 0 to (m_dwTexHeight - 1) do
  begin
    for x:= 0 to (m_dwTexWidth - 1) do
    begin
      Acc:= 0; // clear accumulator
      for i:= -(KERNEL_SIZE div 2) to (KERNEL_SIZE div 2) do
      begin
        pDstRow:= PByte(Integer(pDstData) + d3dlr.Pitch*(y+i));
        for j:= -(KERNEL_SIZE div 2) to (KERNEL_SIZE div 2) do
        begin
          //
          pDst16:= PWord(Integer(pDstRow) + (x+j)*2);
          if (DWORD(pDst16) < DWORD(pDstData)) or (DWORD(pDst16) >= DWORD(pDstMax)) then
          begin
          end else
          begin
            Inc(Acc, (pDst16^ shr 12)); // Increment on Alpha value of texture
          end;
        end;
      end;

      pIntData:= PDWORD(Integer(pSwzData) + (Integer(m_dwTexWidth)*y + x)*4);
      pIntData^:= (Acc and $FF) shl 24; // Store height in alpha bits of ARGB32
    end;
  end;
  // Done updating texture, so clean up used objects
  m_pTexture.UnlockRect(0);
  HeightTexture.UnlockRect(0);

  // Create and lock NormalMap texture
  //todo: Later try to use other (less resource consuming) textures
  Result:= m_pd3dDevice.CreateTexture(m_dwTexWidth, m_dwTexHeight, 1,
                                      0, D3DFMT_A8R8G8B8,
                                      D3DPOOL_MANAGED, m_pNormTexture);
  if FAILED(Result) then Exit;

  // Calculate normal map from intensivity map
  Result:= m_pd3dDevice.CreateTexture(m_dwTexWidth, m_dwTexHeight, 1,
                                      0, D3DFMT_A8R8G8B8,
                                      D3DPOOL_MANAGED, m_pNormTexture);
  if FAILED(Result) then Exit;

  D3DXComputeNormalMap(m_pNormTexture, HeightTexture, nil, 0, D3DX_CHANNEL_ALPHA, 10);

  HeightTexture:= nil; // this step is really unnecessary

  ////////////////////////////////////////////////////////
  // Fill normal texture with alpha mask
  m_pNormTexture.LockRect(0, d3dlr, nil, 0);
  pIntData:= d3dlr.pBits; // pDstData - is 32bit A8R8G8B8 surface

  m_pTexture.LockRect(0, d3dlrHeight, nil, 0);
  pDst16:= d3dlrHeight.pBits; // pDstData - is 32bit A8R8G8B8 surface

  // Calculate per-pixel intensivity of some area
  for y:= 0 to (m_dwTexHeight - 1) do
  begin
    for x:= 0 to (m_dwTexWidth - 1) do
    begin
      if (pDst16^ and ($F shl 12) <> 0)
      then pIntData^:= (pIntData^ or $FF000000)   // set alpha to 0xFF
      else pIntData^:= (pIntData^ and $00FFFFFF); // clear alpha to 0

      Inc(pIntData);
      Inc(pDst16);
    end;
    Inc(pIntData, DWORD(d3dlr.Pitch) - m_dwTexWidth*4);
    Inc(pDst16, DWORD(d3dlrHeight.Pitch) - m_dwTexWidth*2);
  end;

  m_pTexture.UnlockRect(0);
  m_pNormTexture.UnlockRect(0);

  ////////////////////////////////////////////////////////
  m_pTexture:= m_pNormTexture;
  m_pNormTexture:= nil;

  Result:= S_OK;
end;

function TD3DFontDot3.RestoreDeviceObjects: HRESULT;
var
  which: LongWord;
begin
  inherited RestoreDeviceObjects;

  // Create the state blocks for rendering text
  for which:= 0 to 1 do
  begin
    m_pd3dDevice.BeginStateBlock;

    m_pd3dDevice.SetTexture(0, m_pTexture);
    m_pd3dDevice.SetTexture(1, m_pTexture);

    if (D3DFONT_ZENABLE and m_dwFontFlags) <> 0 then
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue)
    else
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iFalse);

    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,         D3DBLEND_SRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE,  iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHAFUNC,        D3DCMP_GREATEREQUAL);
    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE,         D3DFILL_SOLID);
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,         D3DCULL_CCW);
    m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPING,         iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_EDGEANTIALIAS,    iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE,  iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND,      iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,        iFalse);

    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_DOTPRODUCT3);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TFACTOR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DOTPRODUCT3);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_TFACTOR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);

    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAARG2, D3DTA_CURRENT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_NONE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);

    m_pd3dDevice.SetTextureStageState(2, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    m_pd3dDevice.SetTextureStageState(2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    if (which = 0) then
      m_pd3dDevice.EndStateBlock(m_dwSavedStateBlock)
    else
      m_pd3dDevice.EndStateBlock(m_dwDrawTextStateBlock);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DrawText()
// Desc: Draws 2D text
//-----------------------------------------------------------------------------
function TD3DFontDot3.DrawText(x, y: Single; dwColor: DWORD;
  strText: PChar; dwFlags: DWORD): HRESULT;
var
  fStartX: Single;
  pVertices: ^FONT2DVERTEX;
  dwNumTriangles: DWORD;
  c: Char;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  if (m_pd3dDevice = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Setup renderstate
  m_pd3dDevice.CaptureStateBlock(m_dwSavedStateBlock);
  m_pd3dDevice.ApplyStateBlock(m_dwDrawTextStateBlock);
  m_pd3dDevice.SetVertexShader(D3DFVF_FONT2DVERTEX);
  m_pd3dDevice.SetPixelShader(0);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(FONT2DVERTEX));

  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR,
    VectortoRGBA(fNormal, 1));

  // Set filter states
  if (dwFlags and D3DFONT_FILTERED) = D3DFONT_FILTERED then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);

    m_pd3dDevice.SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  end;

  if fModulate2X then
  begin
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_MODULATE2X);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_MODULATE2X);
  end;

  if fOnlyDot3 then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  end;

  if fBlendToBlack then
  begin
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  end;

  fStartX := x;

  // Fill vertex buffer
  pVertices:= nil;
  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  while (strText^ <> #0) do
  begin
    c := strText^;
    Inc(strText);

    if (c = #10) then
    begin
      x := fStartX;
      y := y + (m_fTexCoords[0][3]-m_fTexCoords[0][1])*m_dwTexHeight;
    end;
    if (c < ' ') then Continue;

    tx1 := m_fTexCoords[byte(c)-32][0];
    ty1 := m_fTexCoords[byte(c)-32][1];
    tx2 := m_fTexCoords[byte(c)-32][2];
    ty2 := m_fTexCoords[byte(c)-32][3];

    w := (tx2-tx1) * m_dwTexWidth  / m_fTextScale;
    h := (ty2-ty1) * m_dwTexHeight / m_fTextScale;

    if (c <> ' ') then
    begin
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+h-0.5,0.9,1.0), dwColor, tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+0-0.5,0.9,1.0), dwColor, tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      dwNumTriangles :=dwNumTriangles + 2;

      if dwNumTriangles*3 > (MAX_NUM_VERTICES-6)  then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        pVertices:= nil;
        m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    x := x + w;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

function TD3DFontDot3.SetLightNormal(n: TD3DXVector3): HRESULT;
begin
  D3DXVec3Normalize(fNormal, n);
  Result:= S_OK;
end;

function TD3DFontDot3.DeleteDeviceObjects: HRESULT;
begin
  Result:= inherited DeleteDeviceObjects;
end;

destructor TD3DFontDot3.Destroy;
begin
  InvalidateDeviceObjects;
  DeleteDeviceObjects;
  inherited Destroy;
end;

end.

